@@ -8,11 +8,8 @@ from bigstream.metrics import patch_mutual_information
 from bigstream import features
 import cv2
 
-# TODO: bug! fix_spacing is overwritten after resolve_sampling is called
-#       but downstream functions rely on it having the original value
 
-
-def resolve_sampling(
+def apply_alignment_spacing(
     fix,
     mov,
     fix_mask,
@@ -22,9 +19,51 @@ def resolve_sampling(
     alignment_spacing,
 ):
     """
-    Get mask spacings and skip sample all inputs to alignment_spacing
+    Skip sample all images to as close to alignment_spacing as possible
+    Determine new voxel spacings
+
+    Parameters
+    ----------
+    fix : nd-array
+        The fixed image
+
+    mov : nd-array
+        The moving image
+
+    fix_mask : nd-array
+        The fixed image mask (can be None)
+        Can have a different shape than fix, but assumed to have the same
+        domain or field of view
+
+    mov_mask : nd-array
+        The moving image mask (can be None)
+        Can have a different shape than mov, but assumed to have the same
+        domain or field of view
+
+    fix_spacing : 1d-array
+        The fixed image voxel spacing
+
+    mov_spacing : 1d-array
+        The moving image voxel spacing
+
+    Returns
+    -------
+    Returns 8 values in a tuple
+
+    1. skip sampled fixed image
+    2. skip sampled moving image
+    3. skip sampled fix_mask (or None)
+    4. skip sampled mov_mask (or None)
+    5. spacing of skip sampled fixed image
+    6. spacing of skip sampled moving image
+    7. spacing of skip sampled fixed mask (or None)
+    8. spacing of skip sampled moving mask (or None)
     """
 
+    # ensure spacings are floating point
+    fix_spacing = fix_spacing.astype(np.float64)
+    mov_spacing = mov_spacing.astype(np.float64)
+
     # get mask spacings
     fix_mask_spacing = None
     if fix_mask is not None:
@@ -38,9 +77,13 @@ def resolve_sampling(
         fix, fix_spacing = ut.skip_sample(fix, fix_spacing, alignment_spacing)
         mov, mov_spacing = ut.skip_sample(mov, mov_spacing, alignment_spacing)
         if fix_mask is not None:
-            fix_mask, fix_mask_spacing = ut.skip_sample(fix_mask, fix_mask_spacing, alignment_spacing)
+            fix_mask, fix_mask_spacing = ut.skip_sample(
+                fix_mask, fix_mask_spacing, alignment_spacing,
+            )
         if mov_mask is not None:
-            mov_mask, mov_mask_spacing = ut.skip_sample(mov_mask, mov_mask_spacing, alignment_spacing)
+            mov_mask, mov_mask_spacing = ut.skip_sample(
+                mov_mask, mov_mask_spacing, alignment_spacing,
+            )
 
     return (fix, mov, fix_mask, mov_mask,
             fix_spacing, mov_spacing, fix_mask_spacing, mov_mask_spacing,)
@@ -59,27 +102,118 @@ def images_to_sitk(
     mov_origin,
 ):
     """
-    Convenience function for converting all inputs to sitk images
+    Convert all image inputs to SimpleITK image objects
+
+    Parameters
+    ----------
+    fix : nd-array
+        The fixed image
+
+    mov : nd-array
+        The moving image
+
+    fix_mask : nd-array
+        The fixed image mask (can be None)
+
+    mov_mask : nd-array
+        The moving image mask (can be None)
+
+    fix_spacing : 1d-array
+        The voxel spacing of the fixed image
+
+    mov_spacing : 1d-array
+        The voxel spacing of the moving image
+
+    fix_mask_spacing : 1d-array
+        The voxel spacing of the fixed image mask (can be None)
+        fix and fix_mask are assumed to have the same domain,
+        but this assumption can be slightly broken after skip_sampling
+
+    mov_mask_spacing : 1d-array
+        The voxel spacing of the moving image mask (can be None)
+        mov and mov_mask are assumed to have the same domain,
+        but this assumption can be slightly broken after skip_sampling
+
+    Returns
+    -------
+    Returns 4 values in a tuple
+
+    1. fix image as sitk.Image object
+    2. mov image as sitk.Image object
+    3. fix_mask as sitk.Image object (or None)
+    4. mov_mask as sitk.Image object (or None)
     """
 
-    fix = sitk.Cast(ut.numpy_to_sitk(fix, fix_spacing, origin=fix_origin), sitk.sitkFloat32)
-    mov = sitk.Cast(ut.numpy_to_sitk(mov, mov_spacing, origin=mov_origin), sitk.sitkFloat32)
-    if fix_mask is not None: fix_mask = ut.numpy_to_sitk(fix_mask, fix_mask_spacing, origin=fix_origin)
-    if mov_mask is not None: mov_mask = ut.numpy_to_sitk(mov_mask, mov_mask_spacing, origin=mov_origin)
+    fix = sitk.Cast(ut.numpy_to_sitk(
+        fix, fix_spacing, origin=fix_origin), sitk.sitkFloat32)
+    mov = sitk.Cast(ut.numpy_to_sitk(
+        mov, mov_spacing, origin=mov_origin), sitk.sitkFloat32)
+    if fix_mask is not None:
+        fix_mask = ut.numpy_to_sitk(
+            fix_mask, fix_mask_spacing, origin=fix_origin)
+    if mov_mask is not None:
+        mov_mask = ut.numpy_to_sitk(
+            mov_mask, mov_mask_spacing, origin=mov_origin)
     return fix, mov, fix_mask, mov_mask
 
 
+def format_static_transform_data(
+    transforms,
+    fix,
+    fix_spacing,
+    fix_origin,
+):
+    """
+    Set transform_spacings and transform_origins explicitly
+
+    Parameters
+    ----------
+    transforms : list of nd-arrays
+        The list of static transforms
+
+    fix : nd-array
+        The fixed image
+
+    fix_spacing : 1d-array
+        The voxel spacing of the fixed image
+
+    fix_origin : 1d-array
+        The origin of the fixed image (can be None)
+
+    Returns
+    -------
+    Returns 2 values in a tuple
+
+    1. The tuple of transform spacings
+    2. The tuple of transform origins
+    """
+
+    spacings = []
+    for transform in transforms:
+        spacing = fix_spacing
+        if len(transform.shape) not in [1, 2]:
+            spacing = ut.relative_spacing(transform, fix, fix_spacing)
+        spacings.append(spacing)
+    spacings = tuple(spacings)
+    origins = (fix_origin,)*len(transforms)
+    return (spacings, origins)
+
+
 def feature_point_ransac_affine_align(
     fix, mov,
     fix_spacing,
     mov_spacing,
     blob_sizes,
+    alignment_spacing=None,
     num_sigma_max=15,
     cc_radius=12,
     nspots=5000,
     match_threshold=0.7,
+    max_spot_match_distance=None,
     align_threshold=2.0,
-    diagonal_constraint=0.75,
+    diagonal_constraint=0.25,
+    fix_spot_detection_kwargs={},
+    mov_spot_detection_kwargs={},
     fix_spots=None,
     mov_spots=None,
     fix_mask=None,
@@ -91,9 +225,24 @@ def feature_point_ransac_affine_align(
     **kwargs,
 ):
     """
-    Compute an affine alignment from feature points and ransac. A blob detector finds feature points
-    and a correspondence between those points is estimated. A ransac filter determines the affine
-    transform that brings the largest number of corresponding points to the same locations.
+    Currently this function only works on 3D images.
+
+    Compute an affine alignment from feature points and ransac.
+    A blob detector finds feature points in fix and mov. Correspondence
+    between the fix and mov point sets is estimated using neighborhood
+    correlation. A ransac filter determines the affine transform that brings
+    the largest number of corresponding points to the same locations.
+
+    At least 100 spots must be found in the fixed image and 100 spots
+    in the moving image, otherwise default is returned. At least 50
+    correspondence pairs must be found, otherwise default is returned.
+    These constraints are required for reasonable performance from the
+    ransac affine alignment algorithm.
+
+    If insufficient points are found modify fix_spot_detection_kwargs
+    and/or mov_spot_detection_kwargs. See bigstream.features.
+
+    If insufficient point matches are found, modify match_threshold.
 
     Parameters
     ----------
@@ -125,15 +274,32 @@ def feature_point_ransac_affine_align(
 
     nspots : scalar int (default: 5000)
         The maximum number of feature point spots to use in each image
+        If more spots are found the brightest ones are used.
 
     match_threshold : scalar float in range [0, 1] (default: 0.7)
         The minimum correlation two feature point neighborhoods must have to
         consider them corresponding points
 
-    align_threshold :
+    max_spot_match_distance : scalar float (default: None)
+        The maximum distance a fix and mov spot can be before alignment
+        to still be considered matching spots; in microns. This helps
+        prevent false positive correspondences.
 
-    diagonal_constraint : scalar float (default: 0.75)
-        Diagonal entries of the affine matrix cannot be lower than this number
+    align_threshold : scalar float (default: 2.0)
+        The maximum distance two points can be to be considered aligned
+        by the affine transform; in microns.
+
+    diagonal_constraint : scalar float (default: 0.25)
+        Diagonal entries of the affine matrix cannot be lower than
+        1 - diagonal_contraint or higher than 1 + diagonal_contraint. 
+        If this condition is violated the default transform is returned.
+        This helps prevent bad alignments.
+
+    fix_spot_detection_kwargs : dict (default {})
+        Arguments passed to bigstream.features.blob_detection for fixed image
+
+    mov_spot_detection_kwargs : dict (default {})
+        Arguments passed to bigstream.features.blob_detection for moving image
 
     fix_spots : nd-array Nx3 (default: None)
         Skip the spot detection for the fixed image and provide your own spot coordinate
@@ -147,10 +313,10 @@ def feature_point_ransac_affine_align(
     mov_mask : binary nd-array (default: None)
         Spots from moving image can only be found in the foreground of this mask
 
-    fix_origin : 1d array (default: (0, 0, 0))
+    fix_origin : 1d array (default: all zeros)
         The origin of the fixed image in physical units
 
-    mov_origin : 1d array (default: (0, 0, 0))
+    mov_origin : 1d array (default: all zeros)
         The origin of the moving image in physical units
 
     static_transform_list : list of numpy arrays (default: [])
@@ -159,7 +325,7 @@ def feature_point_ransac_affine_align(
         can be different. I.e. the origin and span are the same (in phyiscal
         units) but the number of voxels can be different.
 
-    default : 2d array 4x4 (default: None)
+    default : 2d array 4x4 (default: identity)
         A default transform to return if the method fails to find a valid one
 
     **kwargs : any additional keyword arguments
@@ -182,55 +348,107 @@ def feature_point_ransac_affine_align(
             fix_origin=fix_origin,
             mov_origin=mov_origin,
         )
+        if mov_mask is not None:
+            mov_mask = apply_transform(
+                fix.astype(mov_mask.dtype), mov_mask,
+                fix_spacing, mov_spacing,
+                transform_list=static_transform_list,
+                fix_origin=fix_origin, 
+                mov_origin=mov_origin,
+                interpolator='0',
+            )
+        mov_spacing = fix_spacing
 
-    # get spots
+    # skip sample and determine mask spacings
+    X = apply_alignment_spacing(
+        fix, mov,
+        fix_mask, mov_mask,
+        fix_spacing, mov_spacing,
+        alignment_spacing,
+    )
+    fix = X[0]
+    mov = X[1]
+    fix_mask = X[2]
+    mov_mask = X[3]
+    fix_spacing = X[4]
+    mov_spacing = X[5]
+    fix_mask_spacing = X[6]
+    mov_mask_spacing = X[7]
+
+    # get fix spots
+    num_sigma = int(min(blob_sizes[1] - blob_sizes[0], num_sigma_max))
+    assert num_sigma > 0, 'num_sigma must be greater than 0, make sure blob_sizes[1] > blob_sizes[0]'
+    print('computing fixed spots', flush=True)
     if fix_spots is None:
+        fix_kwargs = {
+            'num_sigma':num_sigma,
+            'exclude_border':cc_radius,
+            'mask':fix_mask,
+        }
+        fix_kwargs = {**fix_kwargs, **fix_spot_detection_kwargs}
         fix_spots = features.blob_detection(
             fix, blob_sizes[0], blob_sizes[1],
-            num_sigma=min(blob_sizes[1]-blob_sizes[0], num_sigma_max),
-            exclude_border=cc_radius,
+            **fix_kwargs,
         )
+    print(f'found {len(fix_spots)} fixed spots')
+    if len(fix_spots) < 100:
+        print('insufficient fixed spots found, returning default', flush=True)
+        return default
+
+    # get mov spots
+    print('computing moving spots', flush=True)
     if mov_spots is None:
+        mov_kwargs = {
+            'num_sigma':num_sigma,
+            'exclude_border':cc_radius,
+            'mask':mov_mask,
+        }
+        mov_kwargs = {**mov_kwargs, **mov_spot_detection_kwargs}
         mov_spots = features.blob_detection(
             mov, blob_sizes[0], blob_sizes[1],
-            num_sigma=min(blob_sizes[1]-blob_sizes[0], num_sigma_max),
-            exclude_border=cc_radius,
+            **mov_kwargs,
         )
-
-    # TODO: implement masking, remove spots not in foreground
+    print(f'found {len(mov_spots)} moving spots')
+    if len(mov_spots) < 100:
+        print('insufficient moving spots found, returning default', flush=True)
+        return default
 
     # sort
+    print('sorting spots', flush=True)
     sort_idx = np.argsort(fix_spots[:, 3])[::-1]
     fix_spots = fix_spots[sort_idx, :3][:nspots]
     sort_idx = np.argsort(mov_spots[:, 3])[::-1]
     mov_spots = mov_spots[sort_idx, :3][:nspots]
 
     # get contexts
+    print('extracting contexts', flush=True)
     fix_spot_contexts = features.get_contexts(fix, fix_spots, cc_radius)
     mov_spot_contexts = features.get_contexts(mov, mov_spots, cc_radius)
 
+    # get pairwise correlations
+    print('computing pairwise correlations', flush=True)
+    correlations = features.pairwise_correlation(
+        fix_spot_contexts, mov_spot_contexts,
+    )
+
     # convert to physical units
     fix_spots = fix_spots * fix_spacing
     mov_spots = mov_spots * mov_spacing
 
-    # get point correspondences
-    correlations = features.pairwise_correlation(
-        fix_spot_contexts, mov_spot_contexts,
-    )
+    # get matching points
     fix_spots, mov_spots = features.match_points(
         fix_spots, mov_spots,
         correlations, match_threshold,
+        max_distance=max_spot_match_distance,
     )
-
-    # check spot counts
-    if fix_spots.shape[0] < 50:
-        print('Fewer than 50 spots found in fixed image, returning default')
-        return default
-    if mov_spots.shape[0] < 50:
-        print('Fewer than 50 spots found in moving image, returning default')
+    print(f'found {len(fix_spots)} matched spot pairs')
+    if len(fix_spots) < 50 or len(mov_spots) < 50:
+        print('insufficient spot matches found, returning default', flush=True)
         return default
 
     # align
+    # TODO: this is a hard 3D constraint, check opencv for a 2D version
+    print('aligning', flush=True)
     r, Aff, inline = cv2.estimateAffine3D(
         fix_spots, mov_spots,
         ransacThreshold=align_threshold,
@@ -239,13 +457,13 @@ def feature_point_ransac_affine_align(
     )
 
     # ensure affine is sensible
-    if np.any( np.diag(Aff) < diagonal_constraint ):
-        print("Degenerate affine produced, returning default")
+    if np.any( np.abs(np.diag(Aff) - 1) > diagonal_constraint ):
+        print("Degenerate affine produced, returning default", flush=True)
         return default
 
-    # augment to 4x4 matrix and return
-    affine = np.eye(4)
-    affine[:3, :] = Aff
+    # augment matrix and return
+    affine = np.eye(fix.ndim + 1)
+    affine[:fix.ndim, :] = Aff
     return affine
 
 
@@ -271,11 +489,9 @@ def random_affine_search(
     **kwargs,
 ):
     """
-    Apply random affine matrices within given bounds to moving image. The best
-    scoring affines can be further refined with gradient descent based affine
-    alignment. The single best result is returned. This function is intended
-    to find good initialization for a full affine alignment obtained by calling
-    `affine_align`
+    Apply random affine matrices within given bounds to moving image.
+    This function is intended to find good initialization for a full affine
+    alignment obtained by calling `affine_align`
 
     Parameters
     ----------
@@ -377,6 +593,7 @@ def random_affine_search(
             param += (null_value,)
         return param
 
+    # TODO: consider moving to native 2D
     # generalize 2d inputs to 3d
     if fix.ndim == 2:
         fix = fix.reshape(fix.shape + (1,))
@@ -402,8 +619,15 @@ def random_affine_search(
     if max_shear: params[1:, 9:] = F(max_shear)
     center = np.array(fix.shape) / 2 * fix_spacing  # center of rotation
 
+    # format static transform data explicitly
+    a, b = format_static_transform_data(
+        static_transform_list, fix, fix_spacing, fix_origin,
+    )
+    static_transform_spacing = a
+    static_transform_origin = b
+
     # skip sample and determine mask spacings
-    X = resolve_sampling(
+    X = apply_alignment_spacing(
         fix, mov,
         fix_mask, mov_mask,
         fix_spacing, mov_spacing,
@@ -418,17 +642,6 @@ def random_affine_search(
     fix_mask_spacing = X[6]
     mov_mask_spacing = X[7]
 
-    # specify static transform data explicitly
-    static_transform_spacing = []
-    for transform in static_transform_list:
-        spacing = fix_spacing
-        if transform.shape != (4, 4) and len(transform.shape) != 1:
-            spacing = ut.relative_spacing(transform, fix, fix_spacing)
-        static_transform_spacing.append(spacing)
-    static_transform_origin = [fix_origin,]*len(static_transform_list)
-    static_transform_spacing = tuple(static_transform_spacing)
-    static_transform_origin = tuple(static_transform_origin)
-
     # a useful value later, storing prevents redundant function calls
     WORST_POSSIBLE_SCORE = np.finfo(np.float64).max
 
@@ -455,7 +668,7 @@ def random_affine_search(
                     mov_origin=mov_origin,
                     transform_spacing=static_transform_spacing,
                     transform_origin=static_transform_origin,
-                    interpolate_with_nn=True,
+                    interpolator='0',
                 )
             # evaluate metric
             # TODO: this function needs to be updated for different
@@ -502,7 +715,7 @@ def random_affine_search(
     current_best_score = WORST_POSSIBLE_SCORE
     scores = np.empty(random_iterations + 1, dtype=np.float64)
     for iii, ppp in enumerate(params):
-        scores[iii] = score_affine(ut.physical_parameters_to_affine_matrix(ppp, center))
+        scores[iii] = score_affine(ut.physical_parameters_to_affine_matrix_3d(ppp, center))
         if print_running_improvements and scores[iii] < current_best_score:
                 current_best_score = scores[iii]
                 print(iii, ': ', current_best_score, '\n', ppp)
@@ -511,7 +724,7 @@ def random_affine_search(
     # return top results
     partition_indx = np.argpartition(scores, nreturn)[:nreturn]
     params, scores = params[partition_indx], scores[partition_indx]
-    return [ut.physical_parameters_to_affine_matrix(p, center) for p in params[np.argsort(scores)]]
+    return [ut.physical_parameters_to_affine_matrix_3d(p, center) for p in params[np.argsort(scores)]]
 
 
 def affine_align(
@@ -613,19 +826,15 @@ def affine_align(
     if initial_transform_given and np.all(default == np.eye(fix.ndim + 1)):
         default = initial_condition
 
-    # specify static transform data explicitly
-    static_transform_spacing = []
-    for transform in static_transform_list:
-        spacing = fix_spacing
-        if transform.shape != (4, 4) and len(transform.shape) != 1:
-            spacing = ut.relative_spacing(transform, fix, fix_spacing)
-        static_transform_spacing.append(spacing)
-    static_transform_origin = [fix_origin,]*len(static_transform_list)
-    static_transform_spacing = tuple(static_transform_spacing)
-    static_transform_origin = tuple(static_transform_origin)
+    # format static transform data explicitly
+    a, b = format_static_transform_data(
+        static_transform_list, fix, fix_spacing, fix_origin,
+    )
+    static_transform_spacing = a
+    static_transform_origin = b
 
     # skip sample and convert inputs to sitk images
-    X = resolve_sampling(
+    X = apply_alignment_spacing(
         fix, mov,
         fix_mask, mov_mask,
         fix_spacing, mov_spacing,
@@ -649,18 +858,23 @@ def affine_align(
             static_transform_origin,
         )
         irm.SetMovingInitialTransform(T)
+
+    # distinguish between 2D and 3D for rigid transforms
+    ndims = fix.GetDimension()
+    rigid_transform_constructor = sitk.Euler2DTransform if ndims == 2 else sitk.Euler3DTransform
+
     # set transform to optimize
     if isinstance(initial_condition, str) and initial_condition == "CENTER":
         a, b = fix, mov
         if fix_mask is not None and mov_mask is not None:
             a, b = fix_mask, mov_mask
-        x = sitk.CenteredTransformInitializer(a, b, sitk.Euler3DTransform())
-        x = sitk.Euler3DTransform(x).GetTranslation()[::-1]
-        initial_condition = np.eye(4)
-        initial_condition[:3, -1] = x
+        x = sitk.CenteredTransformInitializer(a, b, rigid_transform_constructor())
+        x = rigid_transform_constructor(x).GetTranslation()[::-1]
+        initial_condition = np.eye(ndims+1)
+        initial_condition[:ndims, -1] = x
         initial_transform_given = True
     if rigid and not initial_transform_given:
-        transform = sitk.Euler3DTransform()
+        transform = rigid_transform_constructor()
     elif rigid and initial_transform_given:
         transform = ut.matrix_to_euler_transform(initial_condition)
     elif not rigid and not initial_transform_given:
@@ -740,7 +954,7 @@ def deformable_align(
     control_point_levels : list of type int
         The optimization scales for control point spacing. E.g. if
         `control_point_spacing` is 100.0 and `control_point_levels`
-        is [1, 2, 4] then method will optimize at 400.0 units control
+        is [4, 2, 1] then method will optimize at 400.0 units control
         points spacing, then optimize again at 200.0 units, then again
         at the requested 100.0 units control point spacing.
     
@@ -801,19 +1015,15 @@ def deformable_align(
     initial_fix_shape = fix.shape
     initial_fix_spacing = fix_spacing
 
-    # specify static transform data explicitly
-    static_transform_spacing = []
-    for transform in static_transform_list:
-        spacing = fix_spacing
-        if transform.shape != (4, 4) and len(transform.shape) != 1:
-            spacing = ut.relative_spacing(transform, fix, fix_spacing)
-        static_transform_spacing.append(spacing)
-    static_transform_origin = [fix_origin,]*len(static_transform_list)
-    static_transform_spacing = tuple(static_transform_spacing)
-    static_transform_origin = tuple(static_transform_origin)
+    # format static transform data explicitly
+    a, b = format_static_transform_data(
+        static_transform_list, fix, fix_spacing, fix_origin,
+    )
+    static_transform_spacing = a
+    static_transform_origin = b
 
     # skip sample and convert inputs to sitk images
-    X = resolve_sampling(
+    X = apply_alignment_spacing(
         fix, mov,
         fix_mask, mov_mask,
         fix_spacing, mov_spacing,
@@ -829,15 +1039,17 @@ def deformable_align(
 
     # set up registration object
     irm = configure_irm(**kwargs)
+
     # initial control point grid
-    z = control_point_spacing * control_point_levels[-1]
+    z = control_point_spacing * control_point_levels[0]
     initial_cp_grid = [max(1, int(x*y/z)) for x, y in zip(fix.GetSize(), fix.GetSpacing())]
     transform = sitk.BSplineTransformInitializer(
         image1=fix, transformDomainMeshSize=initial_cp_grid, order=3,
     )
     irm.SetInitialTransformAsBSpline(
-        transform, inPlace=True, scaleFactors=control_point_levels,
+        transform, inPlace=True, scaleFactors=control_point_levels[::-1],
     )
+
     # set initial static transforms
     if static_transform_list:
         T = ut.transform_list_to_composite_transform(
@@ -856,7 +1068,7 @@ def deformable_align(
         field = ut.bspline_to_displacement_field(
             transform, initial_fix_shape,
             spacing=initial_fix_spacing, origin=fix_origin,
-            direction=np.eye(3),
+            direction=np.eye(fix.GetDimension()),
         )
         default = (params, field)
 
@@ -877,7 +1089,7 @@ def deformable_align(
         field = ut.bspline_to_displacement_field(
             transform, initial_fix_shape,
             spacing=initial_fix_spacing, origin=fix_origin,
-            direction=np.eye(3),
+            direction=np.eye(fix.GetDimension()),
         )
         print("Registration succeeded", flush=True)
         return params, field
@@ -998,14 +1210,13 @@ def alignment_pipeline(
              'deform':lambda **c: deformable_align(*a, **{**b, **c})[1],}
 
     # loop over steps
-    initial_transform_count = len(static_transform_list)
+    new_transforms = []
     for alignment, arguments in steps:
         arguments = {**kwargs, **arguments}
-        arguments['static_transform_list'] = static_transform_list
-        static_transform_list.append(align[alignment](**arguments))
+        arguments['static_transform_list'] = static_transform_list + new_transforms
+        new_transforms.append(align[alignment](**arguments))
 
     # return in the requested format
-    new_transforms = static_transform_list[initial_transform_count:]
     if return_format == 'independent':
         return new_transforms
     elif return_format == 'compressed':
