@@ -1,12 +1,17 @@
 import numpy as np
-from fishspot.filter import white_tophat
+from fishspot.filter import white_tophat, apply_foreground_mask
 from fishspot.detect import detect_spots_log
+from scipy.stats.mstats import winsorize
+from scipy.spatial import cKDTree
 
 
 def blob_detection(
     image,
     min_blob_radius,
     max_blob_radius,
+    winsorize_limits=None,
+    background_subtract=False,
+    mask=None,
     **kwargs,
 ):
     """
@@ -20,6 +25,11 @@ def blob_detection(
         The smallest size blob you want to find in voxel units
     max_blob_radius : scalar float
         The largest size blob you want to find in voxel units
+    winsorize_limits : tuple of two floats (default: None)
+        If not None, winsorize the input with (min, max) percentile cutoffs
+    background_subtract : bool (default: False)
+        If True, use white_tophat background subtraction with max_blob_radius
+        as the filter radius
     **kwargs : any additional kwargs
         Passed to fishspot.detect_spots_log
 
@@ -31,15 +41,20 @@ def blob_detection(
         detected coordinate location.
     """
 
-    wth = white_tophat(image, max_blob_radius)
+    processed_image = np.copy(image)
+    if winsorize_limits is not None:
+        processed_image = winsorize(processed_image, limits=winsorize_limits)
+    if background_subtract:
+        processed_image = white_tophat(processed_image, max_blob_radius)
     spots = detect_spots_log(
-        wth,
+        processed_image,
         min_blob_radius,
         max_blob_radius,
         **kwargs,
     ).astype(int)
-    intensities = image[spots[:, 0], spots[:, 1], spots[:, 2]]
-    return np.hstack((spots[:, :3], intensities[..., None]))
+    if mask is not None: spots = apply_foreground_mask(spots, mask)
+    intensities = image[ tuple(spots[:, iii] for iii in range(image.ndim)) ]
+    return np.hstack((spots[:, :image.ndim], intensities[..., None]))
 
 
 def get_contexts(image, coords, radius):
@@ -60,10 +75,12 @@ def get_contexts(image, coords, radius):
     neighborhoods : list of nd-arrays
         List of the extracted neighborhoods
     """
-
+    if isinstance(radius, (int, np.integer)):
+        radius = (radius,) * image.ndim  # Convert scalar radius to ndim radius
+ 
     contexts = []
     for coord in coords:
-        crop = tuple(slice(x-radius, x+radius+1) for x in coord)
+        crop = tuple(slice(int(x - r), int(x + r + 1)) for x, r in zip(coord, radius))
         contexts.append(image[crop])
     return contexts    
 
@@ -118,7 +135,7 @@ def pairwise_correlation(A, B):
     return corr
 
 
-def match_points(a_pos, b_pos, scores, threshold):
+def match_points(a_pos, b_pos, scores, threshold, max_distance=None):
     """
     Given two point sets and pairwise scores, determine which points correspond.
 
@@ -132,6 +149,8 @@ def match_points(a_pos, b_pos, scores, threshold):
         Correspondence scores for all points in a_pos to all points in b_pos
     threshold : scalar float
         Minimum correspondence score for a valid match
+    max_distance : float (default: None)
+        The maximum distance two spots can be and still be matched
 
     Returns
     -------
@@ -139,6 +158,17 @@ def match_points(a_pos, b_pos, scores, threshold):
         The points from a_pos and b_pos that correspond
     """
 
+    # only points within max_distance should be considered
+    max_score = np.max(scores) + 1
+    if max_distance is not None:
+        a_kdtree = cKDTree(a_pos)
+        valid_pairs = a_kdtree.query_ball_tree(
+            cKDTree(b_pos), max_distance,
+        )
+        for iii, fancy_index in enumerate(valid_pairs):
+            scores[iii, fancy_index] += max_score
+        threshold += max_score
+
     # get highest scores above threshold
     best_indcs = np.argmax(scores, axis=1)
     a_indcs = range(len(a_pos))
