@@ -7,9 +7,10 @@ from dask.distributed import as_completed, wait
 from ClusterWrap.decorator import cluster
 import bigstream.utility as ut
 from bigstream.align import alignment_pipeline
-from bigstream.transform import apply_transform
+from bigstream.transform import apply_transform, compose_transform_list
 from bigstream.transform import apply_transform_to_coordinates
 from bigstream.transform import compose_transforms
+from distributed import Lock, MultiLock
 
 
 @cluster
@@ -25,6 +26,7 @@ def distributed_piecewise_alignment_pipeline(
     mov_mask=None,
     foreground_percentage=0.5,
     static_transform_list=[],
+    rebalance_for_missing_neighbors=True,
     cluster=None,
     cluster_kwargs={},
     temporary_directory=None,
@@ -96,6 +98,13 @@ def distributed_piecewise_alignment_pipeline(
         can be different. I.e. the origin and span are the same (in physical
         units) but the number of voxels can be different.
 
+    rebalance_for_missing_neighbors : bool (default: True)
+        If True, when a block has a missing neighbor, it's linear blending weights
+        are rebalanced to account for the missing neighbor. The ensures transforms
+        aren't artificially dampened by missing neighbors, which have zero valued
+        displacement. However, if it is desired that the edges of masked regions
+        should more smoothly dampen into backgroun, then set this to false.
+
     cluster : ClusterWrap.cluster object (default: None)
         Only set if you have constructed your own static cluster. The default behavior
         is to construct a cluster for the duration of this function, then close it
@@ -135,25 +144,32 @@ def distributed_piecewise_alignment_pipeline(
         the displacement vector.
     """
 
+    ################  Input, output, and parameter formatting ############
     # temporary file paths and create zarr images
     temporary_directory = tempfile.TemporaryDirectory(
         prefix='.', dir=temporary_directory or os.getcwd(),
     )
+    zarr_blocks = (128,) * fix.ndim
     fix_zarr_path = temporary_directory.name + '/fix.zarr'
     mov_zarr_path = temporary_directory.name + '/mov.zarr'
     fix_mask_zarr_path = temporary_directory.name + '/fix_mask.zarr'
     mov_mask_zarr_path = temporary_directory.name + '/mov_mask.zarr'
-    fix_zarr = ut.numpy_to_zarr(fix, blocksize, fix_zarr_path)
-    mov_zarr = ut.numpy_to_zarr(mov, blocksize, mov_zarr_path)
-    if fix_mask is not None: fix_mask_zarr = ut.numpy_to_zarr(fix_mask, blocksize, fix_mask_zarr_path)
-    if mov_mask is not None: mov_mask_zarr = ut.numpy_to_zarr(mov_mask, blocksize, mov_mask_zarr_path)
+    fix_zarr = ut.numpy_to_zarr(fix, zarr_blocks, fix_zarr_path)
+    mov_zarr = ut.numpy_to_zarr(mov, zarr_blocks, mov_zarr_path)
+    fix_mask_zarr = None
+    if fix_mask is not None:
+        fix_mask_zarr = ut.numpy_to_zarr(fix_mask, zarr_blocks, fix_mask_zarr_path)
+    mov_mask_zarr = None
+    if mov_mask is not None:
+        mov_mask_zarr = ut.numpy_to_zarr(mov_mask, zarr_blocks, mov_mask_zarr_path)
 
     # zarr files for initial deformations
     new_list = []
+    zarr_blocks = (128,) * fix.ndim + (fix.ndim,)
     for iii, transform in enumerate(static_transform_list):
-        if transform.shape != (4, 4) and len(transform.shape) != 1:
+        if len(transform.shape) > fix.ndim:  # fields are always fix.ndim + 1
             path = temporary_directory.name + f'/deform{iii}.zarr'
-            transform = ut.numpy_to_zarr(transform, tuple(blocksize) + (transform.shape[-1],), path)
+            transform = ut.numpy_to_zarr(transform, zarr_blocks, path)
         new_list.append(transform)
     static_transform_list = new_list
 
@@ -165,10 +181,18 @@ def distributed_piecewise_alignment_pipeline(
             tuple(blocksize) + (fix.ndim,),
             np.float32,
         )
+        # chunks will need locks for parallel writing
+        locks = [Lock(f'{x}') for x in np.ndindex(*output_transform.cdata_shape)]
 
+    # merge keyword arguments with step specific keyword arguments
+    steps = [(a, {**kwargs, **b}) for a, b in steps]
+    ######################################################################
+
+
+    #################### Determine block structure #######################
     # determine fixed image slices for blocking
     blocksize = np.array(blocksize)
-    nblocks = np.ceil(np.array(fix.shape) / blocksize).astype(int)
+    nblocks = np.ceil(fix.shape / blocksize).astype(int)
     overlaps = np.round(blocksize * overlap).astype(int)
     indices, slices = [], []
     for (i, j, k) in np.ndindex(*nblocks):
@@ -194,54 +218,51 @@ def distributed_piecewise_alignment_pipeline(
             slices.append(coords)
 
     # determine foreground neighbor structure
-    new_indices = []
-    neighbor_offsets = np.array(list(product([-1, 0, 1], repeat=3)))
-    for index, coords in zip(indices, slices):
-        neighbor_flags = {tuple(o): tuple(index + o) in indices for o in neighbor_offsets}
-        new_indices.append((index, coords, neighbor_flags))
-    indices = new_indices
+    neighbor_flags = []
+    neighbor_offsets = np.array(list(product([-1, 0, 1], repeat=fix.ndim)))
+    for index in indices:
+        flags = {tuple(o): tuple(index + o) in indices for o in neighbor_offsets}
+        neighbor_flags.append(flags)
 
-    # establish all keyword arguments
-    steps = [(a, {**kwargs, **b}) for a, b in steps]
+    # bundle the three parallel lists together
+    block_data = list(zip(indices, slices, neighbor_flags))
+    ######################################################################
 
-    # closure for alignment pipeline
-    def align_single_block(
-        indices,
-        static_transform_list,
-    ):
 
-        # print some feedback
-        print("Block index: ", indices[0], "\nSlices: ", indices[1], flush=True)
+    #################### The function to run on each block ###############
+    def align_single_block(indices, static_transform_list):
 
-        # get the coordinates, read fixed data
+        # parse input, log index and slices
         block_index, fix_slices, neighbor_flags = indices
-        fix = fix_zarr[fix_slices]
+        print("Block index: ", block_index, "\nSlices: ", fix_slices, flush=True)
 
+
+        ########## Map fix block corners onto mov coordinates ############
+        ############## Reads static transforms in the meantime ###########
         # get fixed image block corners in physical units
         fix_block_coords = []
-        for corner in list(product([0, 1], repeat=3)):
+        for corner in list(product([0, 1], repeat=len(fix_slices))):
             a = [x.stop-1 if y else x.start for x, y in zip(fix_slices, corner)]
             fix_block_coords.append(a)
         fix_block_coords = np.array(fix_block_coords)
         fix_block_coords_phys = fix_block_coords * fix_spacing
 
-        # parse initial transforms
-        # recenter affines, read deforms, apply transforms to crop coordinates
+        # read static transforms: recenter affines, apply to crop coordinates
         new_list = []
         mov_block_coords_phys = np.copy(fix_block_coords_phys)
         for transform in static_transform_list[::-1]:
-            if transform.shape == (4, 4):
+            if len(transform.shape) == 2:
                 mov_block_coords_phys = apply_transform_to_coordinates(
                     mov_block_coords_phys, [transform,],
                 )
                 transform = ut.change_affine_matrix_origin(transform, fix_block_coords_phys[0])
             else:
+                spacing = ut.relative_spacing(transform, fix_zarr, fix_spacing)
                 ratio = np.array(transform.shape[:-1]) / fix_zarr.shape
                 start = np.round( ratio * fix_block_coords[0] ).astype(int)
                 stop = np.round( ratio * (fix_block_coords[-1] + 1) ).astype(int)
                 transform_slices = tuple(slice(a, b) for a, b in zip(start, stop))
                 transform = transform[transform_slices]
-                spacing = ut.relative_spacing(transform, fix, fix_spacing)
                 origin = spacing * start
                 mov_block_coords_phys = apply_transform_to_coordinates(
                     mov_block_coords_phys, [transform,], spacing, origin
@@ -249,33 +270,39 @@ def distributed_piecewise_alignment_pipeline(
             new_list.append(transform)
         static_transform_list = new_list[::-1]
 
-        # get moving image crop, read moving data 
+        # Now we can determine the moving image crop
         mov_block_coords = np.round(mov_block_coords_phys / mov_spacing).astype(int)
         mov_start = np.min(mov_block_coords, axis=0)
         mov_stop = np.max(mov_block_coords, axis=0)
         mov_start = np.maximum(0, mov_start)
         mov_stop = np.minimum(np.array(mov_zarr.shape)-1, mov_stop)
         mov_slices = tuple(slice(a, b) for a, b in zip(mov_start, mov_stop))
-        mov = mov_zarr[mov_slices]
 
-        # read masks
+        # get moving crop origin relative to fixed crop
+        mov_origin = mov_start * mov_spacing - fix_block_coords_phys[0]
+        ##################################################################
+
+
+        ################ Read fix and moving data ########################
+        fix = fix_zarr[fix_slices]
+        mov = mov_zarr[mov_slices]
         fix_mask, mov_mask = None, None
-        if os.path.isdir(fix_mask_zarr_path):
+        if fix_mask_zarr is not None:
             ratio = np.array(fix_mask_zarr.shape) / fix_zarr.shape
             start = np.round( ratio * fix_block_coords[0] ).astype(int)
             stop = np.round( ratio * (fix_block_coords[-1] + 1) ).astype(int)
             fix_mask_slices = tuple(slice(a, b) for a, b in zip(start, stop))
             fix_mask = fix_mask_zarr[fix_mask_slices]
-        if os.path.isdir(mov_mask_zarr_path):
+        if mov_mask_zarr is not None:
             ratio = np.array(mov_mask_zarr.shape) / mov_zarr.shape
             start = np.round( ratio * mov_start ).astype(int)
             stop = np.round( ratio * mov_stop ).astype(int)
             mov_mask_slices = tuple(slice(a, b) for a, b in zip(start, stop))
             mov_mask = mov_mask_zarr[mov_mask_slices]
+        ##################################################################
 
-        # get moving image origin
-        mov_origin = mov_start * mov_spacing - fix_block_coords_phys[0]
 
+        ############################ Align ###############################
         # run alignment pipeline
         transform = alignment_pipeline(
             fix, mov, fix_spacing, mov_spacing, steps,
@@ -285,18 +312,21 @@ def distributed_piecewise_alignment_pipeline(
         )
 
         # ensure transform is a vector field
-        if transform.shape == (4, 4):
+        if len(transform.shape) == 2:
             transform = ut.matrix_to_displacement_field(
                 transform, fix.shape, spacing=fix_spacing,
             )
+        ##################################################################
 
+
+        ################ Apply weights for linear blending ###############
         # create the standard weight array
         core = tuple(x - 2*y + 2 for x, y in zip(blocksize, overlaps))
         pad = tuple((2*y - 1, 2*y - 1) for y in overlaps)
         weights = np.pad(np.ones(core, dtype=np.float64), pad, mode='linear_ramp')
 
         # rebalance if any neighbors are missing
-        if not np.all(list(neighbor_flags.values())):
+        if rebalance_for_missing_neighbors and not np.all(list(neighbor_flags.values())):
 
             # define overlap slices
             slices = {}
@@ -317,8 +347,8 @@ def distributed_piecewise_alignment_pipeline(
             weights = weights.astype(np.float32)
 
         # crop weights if block is on edge of domain
-        for i in range(3):
-            region = [slice(None),]*3
+        for i in range(fix.ndim):
+            region = [slice(None),]*fix.ndim
             if block_index[i] == 0:
                 region[i] = slice(overlaps[i], None)
                 weights = weights[tuple(region)]
@@ -328,29 +358,43 @@ def distributed_piecewise_alignment_pipeline(
 
         # crop any incomplete blocks (on the ends)
         if np.any( weights.shape != transform.shape[:-1] ):
-            crop = tuple(slice(0, s) for s in transform.shape[:-1])
-            weights = weights[crop]
+            weights = weights[tuple(slice(0, s) for s in transform.shape[:-1])]
 
         # apply weights
         transform = transform * weights[..., None]
+        ##################################################################
 
-        # return or write the data
+
+        ################ Write or return result ##########################
+        # if there's no write path, just return the transform block
         if not write_path:
             return transform
-        else:
-            # wait until the correct write window for this write group
-            # TODO: if a worker can query the set of running tasks, I may be able to skip
-            #       groups that are completely written
-            write_group = np.sum(np.array(block_index) % 3 * (9, 3, 1))
-            while not (write_group < time.time() / write_group_interval % 27 < write_group + .5):
-                time.sleep(1)
-            output_transform[fix_slices] = output_transform[fix_slices] + transform
-            return True
-    # END CLOSURE
 
+        # otherwise, coordinate with neighboring blocks
+        else:
+            # block until locks for all write blocks are acquired
+            lock_strs = []
+            for delta in product((-1, 0, 1,), repeat=3):
+                lock_strs.append(str(tuple(a + b for a, b in zip(block_index, delta))))
+            lock = MultiLock(lock_strs)
+            lock.acquire()
+
+            # write result to disk
+            print(f'WRITING BLOCK {block_index} at {time.ctime(time.time())}', flush=True)
+            output_transform[fix_slices] = output_transform[fix_slices] + transform
+            print(f'FINISHED WRITING BLOCK {block_index} at {time.ctime(time.time())}', flush=True)
+
+            # release the lock
+            lock.release()
+            return True
+        ##################################################################
+    ######################################################################
+
+
+    #################### Submit all blocks, parse output #################
     # submit all alignments to cluster
     futures = cluster.client.map(
-        align_single_block, indices,
+        align_single_block, block_data,
         static_transform_list=static_transform_list,
     )
 
@@ -361,14 +405,15 @@ def distributed_piecewise_alignment_pipeline(
         for batch in as_completed(futures, with_results=True).batches():
             for future, result in batch:
                 iii = future_keys.index(future.key)
-                transform[indices[iii][1]] += result
+                transform[block_data[iii][1]] += result
         return transform
     else:
         all_written = np.all( cluster.client.gather(futures) )
         return output_transform
+    ######################################################################
 
 
-# TODO: this function not yet refactored
+# TODO: THIS FUNCTION CURRENTLY DOES NOT WORK FOR LARGER THAN MEMORY TRANSFORMS
 @cluster
 def nested_distributed_piecewise_alignment_pipeline(
     fix,
@@ -381,10 +426,14 @@ def nested_distributed_piecewise_alignment_pipeline(
     mov_mask=None,
     cluster=None,
     cluster_kwargs={},
+    temporary_directory=None,
+    write_path=None,
     **kwargs,
 ):
     """
     Compose multiple calls to distributed_piecewise_alignment_pipeline with one function call.
+    Be sure you understand how distributed_piecewise_alignment_pipeline and all its arguments
+    work before you use this function.
 
     Parameters
     ----------
@@ -408,9 +457,11 @@ def nested_distributed_piecewise_alignment_pipeline(
     schedule : list of tuples in this form [(tuple, [(str, dict), (str, dict), ...]), ...]
         The blocksize and steps arguments for each call to distributed_piecewise_alignment_pipeline
         An example:
-            step1 = ( (3, 1, 1), [('affine', affine_kwargs)] )
-            step2 = ( (
-
+            step1 = ( (256, 256, 256), [('affine', affine_kwargs)] )
+            step2 = ( (128, 128, 128), [('affine', affine_kwargs), ('deform', deform_kwargs)] )
+            schedule = [step1, step2]
+        In this example, affine alignment will be done on all 256^3 blocks of the image.
+        Subsequently, affine + deformable alignment will be done on all 128^3 blocks of the image.
 
     static_transform_list : list of numpy arrays (default: [])
         Transforms applied to moving image before applying query transform
@@ -442,130 +493,73 @@ def nested_distributed_piecewise_alignment_pipeline(
         this will be ClusterWrap.local_cluster.
         This is how distribution parameters are specified.
 
+    temporary_directory : string (default: None)
+        Temporary files are created during alignment. The temporary files will be
+        in their own folder within the `temporary_directory`. The default is the
+        current directory. Temporary files are removed if the function completes
+        successfully.
+
+    write_path : string (default: None)
+        If the transforms found by this function are too large to fit into main
+        process memory, set this parameter to a folder where the transforms
+        can be written to disk as separate zarr files
+
     kwargs : any additional arguments
         Passed to `distributed_piecewise_alignment_pipeline`
 
     Returns
     -------
-    field : ndarray
+    field : nd array or zarr.core.Array
+        Composition of all alignments into a single displacement vector field.
     """
 
+    # temporary file paths and create zarr images
+    temporary_directory = tempfile.TemporaryDirectory(
+        prefix='.', dir=temporary_directory or os.getcwd(),
+    )
+    zarr_blocks = [128,] * fix.ndim
+    fix_zarr_path = temporary_directory.name + '/fix.zarr'
+    mov_zarr_path = temporary_directory.name + '/mov.zarr'
+    fix_mask_zarr_path = temporary_directory.name + '/fix_mask.zarr'
+    mov_mask_zarr_path = temporary_directory.name + '/mov_mask.zarr'
+    fix_zarr = ut.numpy_to_zarr(fix, zarr_blocks, fix_zarr_path)
+    mov_zarr = ut.numpy_to_zarr(mov, zarr_blocks, mov_zarr_path)
+    fix_mask_zarr = None
+    if fix_mask is not None: fix_mask_zarr = ut.numpy_to_zarr(fix_mask, zarr_blocks, fix_mask_zarr_path)
+    mov_mask_zarr = None
+    if mov_mask is not None: mov_mask_zarr = ut.numpy_to_zarr(mov_mask, zarr_blocks, mov_mask_zarr_path)
 
+    # zarr files for initial deformations
+    new_list = []
+    for iii, transform in enumerate(static_transform_list):
+        if transform.shape != (4, 4) and len(transform.shape) != 1:
+            path = temporary_directory.name + f'/deform{iii}.zarr'
+            transform = ut.numpy_to_zarr(transform, tuple(zarr_blocks) + (transform.shape[-1],), path)
+        new_list.append(transform)
+    static_transform_list = new_list
 
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-def old()
-
-    # set working copies of moving data
-    if static_transform_list is not None:
-        current_moving = apply_transform(
-            fix, mov, fix_spacing, mov_spacing,
-            transform_list=static_transform_list,
+    # loop over the schedule
+    for iii, (blocksize, steps) in enumerate(schedule):
+        local_write_path = None
+        if write_path: local_write_path = write_path + '/{iii}.zarr'
+        deform = distributed_piecewise_alignment_pipeline(
+            fix_zarr, mov_zarr, fix_spacing, mov_spacing,
+            steps, blocksize,
+            static_transform_list=static_transform_list,
+            fix_mask=fix_mask_zarr,
+            mov_mask=mov_mask_zarr,
+            write_path=local_write_path,
+            cluster=cluster,
+            **kwargs,
         )
-        current_moving_mask = None
-        if mov_mask is not None:
-            current_moving_mask = apply_transform(
-                fix, mov_mask, fix_spacing, mov_spacing,
-                transform_list=static_transform_list,
+        # TODO: THIS DOES NOT WORK WITH LARGER THAN MEMORY TRANSFORMS
+        if iii > 0:
+            deform = compose_transforms(
+                static_transform_list.pop(), deform,
+                fix_spacing, fix_spacing,
             )
-            current_moving_mask = (current_moving_mask > 0).astype(np.uint8)
-    else:
-        current_moving = np.copy(mov)
-        current_moving_mask = None if mov_mask is None else np.copy(mov_mask)
+        static_transform_list.append(deform)
 
-    # initialize container and Loop over outer levels
-    counter = 0  # count each call to distributed_piecewise_affine_align
-    deform = np.zeros(fix.shape + (3,), dtype=np.float32)
-    for outer_level, inner_list in enumerate(block_schedule):
-
-        # initialize inner container and Loop over inner levels
-        ddd = np.zeros_like(deform)
-        for inner_level, nblocks in enumerate(inner_list):
-
-            # determine parameter settings
-            if parameter_schedule is not None:
-                instance_kwargs = {**kwargs, **parameter_schedule[counter]}
-            else:
-                instance_kwargs = kwargs
-
-            # align
-            ddd += distributed_piecewise_alignment_pipeline(
-                fix, current_moving,
-                fix_spacing, fix_spacing,  # images should be on same grid
-                nblocks=nblocks,
-                fix_mask=fix_mask,
-                mov_mask=current_moving_mask,
-                cluster=cluster,
-                cluster_kwargs=cluster_kwargs,
-                **instance_kwargs,
-            )
-
-            # increment counter
-            counter += 1
-
-        # take mean
-        ddd = ddd / len(inner_list)
-
-        # if not first iteration, compose with existing deform
-        if outer_level > 0:
-            deform = compose_transforms(deform, ddd, fix_spacing,)
-        else:
-            deform = ddd
-
-        # combine with initial transforms if given
-        if static_transform_list is not None:
-            transform_list = static_transform_list + [deform,]
-        else:
-            transform_list = [deform,]
-
-        # update working copy of image
-        current_moving = apply_transform(
-            fix, mov, fix_spacing, mov_spacing,
-            transform_list=transform_list,
-        )
-        # update working copy of mask
-        if mov_mask is not None:
-            current_moving_mask = apply_transform(
-                fix, mov_mask, fix_spacing, mov_spacing,
-                transform_list=transform_list,
-            )
-            current_moving_mask = (current_moving_mask > 0).astype(np.uint8)
-
-        # write intermediates
-        if intermediates_path is not None:
-            ois = str(outer_level)
-            deform_path = (intermediates_path + '/twist_deform_{}.npy').format(ois)
-            image_path = (intermediates_path + '/twist_image_{}.npy').format(ois)
-            mask_path = (intermediates_path + '/twist_mask_{}.npy').format(ois)
-            np.save(deform_path, deform)
-            np.save(image_path, current_moving)
-            if mov_mask is not None:
-                np.save(mask_path, current_moving_mask)
-
-    # return deform
-    return deform
-    
+    # return in the requested format
+    return static_transform_list.pop()
 
