@@ -5,7 +5,7 @@ import os, tempfile
 from ClusterWrap.decorator import cluster
 import dask.array as da
 import zarr
-import bigstream.transform as cs_transform
+import bigstream.transform as bs_transform
 
 
 @cluster
@@ -16,7 +16,6 @@ def distributed_apply_transform(
     blocksize,
     write_path=None,
     overlap=0.5,
-    inverse_transforms=False,
     dataset_path=None,
     temporary_directory=None,
     cluster=None,
@@ -57,9 +56,6 @@ def distributed_apply_transform(
     overlap : float in range [0, 1] (default: 0.5)
         Block overlap size as a percentage of block size
 
-    inverse_transforms : bool (default: False)
-        Set to true if the list of transforms are all inverted
-
     dataset_path : string (default: None)
         A subpath in the zarr array to write the resampled data to
 
@@ -141,7 +137,6 @@ def distributed_apply_transform(
         # read relevant region of transforms
         new_list = []
         transform_origin = [fix_origin,] * len(transform_list)
-        shape = fix_zarr.shape if not inverse_transforms else mov_zarr.shape
         for iii, transform in enumerate(transform_list):
             if transform.shape != (4, 4):
                 start = np.floor(fix_origin / kwargs['transform_spacing'][iii]).astype(int)
@@ -159,7 +154,7 @@ def distributed_apply_transform(
             a = [x.stop-1 if y else x.start for x, y in zip(fix_slices, corner)]
             fix_block_coords.append(a)
         fix_block_coords = np.array(fix_block_coords) * fix_spacing
-        mov_block_coords = cs_transform.apply_transform_to_coordinates(
+        mov_block_coords = bs_transform.apply_transform_to_coordinates(
             fix_block_coords, transform_list, kwargs['transform_spacing'], transform_origin,
         )
         mov_block_coords = np.round(mov_block_coords / mov_spacing).astype(int)
@@ -172,7 +167,7 @@ def distributed_apply_transform(
         mov_origin = mov_spacing * [s.start for s in mov_slices]
 
         # resample
-        aligned = cs_transform.apply_transform(
+        aligned = bs_transform.apply_transform(
             fix, mov, fix_spacing, mov_spacing,
             transform_list=transform_list,
             transform_origin=transform_origin,
@@ -308,14 +303,17 @@ def distributed_apply_transform_to_coordinates(
     origin = np.min(coordinates, axis=0)
     nblocks = np.max(coordinates, axis=0) - origin
     nblocks = np.ceil(nblocks / partition_size).astype(int)
-    partitions = []
+    partitions, indices = [], []
     for (i, j, k) in np.ndindex(*nblocks):
         lower_bound = origin + partition_size * np.array((i, j, k))
         upper_bound = lower_bound + partition_size
         not_too_low = np.all(coordinates >= lower_bound, axis=1)
         not_too_high = np.all(coordinates < upper_bound, axis=1)
-        coords = coordinates[ not_too_low * not_too_high ]
-        if coords.size != 0: partitions.append(coords)
+        part_indices = np.nonzero( not_too_low * not_too_high )[0]
+        if part_indices.size != 0:
+            partitions.append(coordinates[part_indices])
+            indices.append(part_indices)
+    indices = np.concatenate(indices, axis=0)
 
     def transform_partition(coordinates, transform_list):
 
@@ -335,7 +333,7 @@ def distributed_apply_transform_to_coordinates(
         transform_list = new_list
 
         # apply transforms
-        return cs_transform.apply_transform_to_coordinates(
+        return bs_transform.apply_transform_to_coordinates(
             coordinates, transform_list,
             transform_spacing,
             transform_origin=a,
@@ -346,9 +344,10 @@ def distributed_apply_transform_to_coordinates(
         transform_partition, partitions,
         transform_list=transform_list,
     )
-    results = cluster.client.gather(futures)
-
-    return np.concatenate(results, axis=0)
+    permuted = np.concatenate(cluster.client.gather(futures), axis=0)
+    results = np.empty_like(permuted)
+    results[indices] = permuted
+    return results
 
 
 @cluster
@@ -357,8 +356,10 @@ def distributed_invert_displacement_vector_field(
     spacing,
     blocksize,
     write_path,
+    step=0.5,
     iterations=10,
-    order=2,
+    sqrt_order=2,
+    sqrt_step=0.5,
     sqrt_iterations=5,
     cluster=None,
     cluster_kwargs={},
@@ -380,13 +381,19 @@ def distributed_invert_displacement_vector_field(
     write_path : string (default: None)
         Location on disk to write the resampled data as a zarr array
 
+    step : float (default: 0.5)
+        The step size used for each iteration of the stationary point algorithm
+
     iterations : scalar int (default: 10)
         The number of stationary point iterations to find inverse. More
         iterations gives a more accurate inverse but takes more time.
 
-    order : scalar int (default: 2)
+    sqrt_order : scalar int (default: 2)
         The number of roots to take before stationary point iterations.
 
+    sqrt_step : float (default: 0.5)
+        The step size used for each iteration of the composition square root gradient descent
+
     sqrt_iterations : scalar int (default: 5)
         The number of iterations to find the field composition square root.
 
@@ -413,6 +420,7 @@ def distributed_invert_displacement_vector_field(
     overlap = np.round(blocksize * 0.25).astype(int)
     nblocks = np.ceil(np.array(field_zarr.shape[:-1]) / blocksize).astype(int)
 
+    # TODO: eliminate use of dask array
     # store block coordinates in a dask array
     block_coords = np.empty(nblocks, dtype=tuple)
     for (i, j, k) in np.ndindex(*nblocks):
@@ -429,10 +437,16 @@ def distributed_invert_displacement_vector_field(
 
         slices = slices.item()
         field = field_zarr[slices]
-        inverse = cs_transform.invert_displacement_vector_field(
-            field, spacing, iterations, order, sqrt_iterations,
+        inverse = bs_transform.invert_displacement_vector_field(
+            field,
+            spacing,
+            step=step,
+            iterations=iterations,
+            sqrt_order=sqrt_order,
+            sqrt_step=sqrt_step,
+            sqrt_iterations=sqrt_iterations,
         )
-        
+
         # crop out overlap
         for axis in range(inverse.ndim - 1):
 
