@@ -1,7 +1,8 @@
 import numpy as np
 import SimpleITK as sitk
 import bigstream.utility as ut
-import os, psutil
+from bigstream.configure_irm import interpolator_switch
+import os
 from scipy.ndimage import map_coordinates
 
 
@@ -13,7 +14,7 @@ def apply_transform(
     transform_origin=None,
     fix_origin=None,
     mov_origin=None,
-    interpolate_with_nn=False,
+    interpolator='1',
     extrapolate_with_nn=False,
 ):
     """
@@ -24,6 +25,7 @@ def apply_transform(
     ----------
     fix : ndarray
         the fixed image
+        Optionally, this can be a tuple specifying a shape.
 
     mov : ndarray
         the moving image; `fix.ndim` must equal `mov.ndim`
@@ -37,7 +39,7 @@ def apply_transform(
         of the moving image. Length must equal `mov.ndim`.
 
     transform_list : list
-        The list of transforms to apply. These may be 2d arrays of shape 4x4
+        The list of transforms to apply. These may be 2d arrays of shape 3x3 or 4x4
         (affine transforms), or ndarrays of `fix.ndim` + 1 dimension (deformations).
         Zarr arrays work just fine.
 
@@ -53,7 +55,7 @@ def apply_transform(
     transform_origin : None (default), 1d array, or tuple of 1d arrays
         The origin in physical units (e.g. mm or um) of the given transforms.
         If None, all origins are assumed to be (0, 0, 0, ...); otherwise, follows
-        the same logic as transform_spacing.
+        the same format as transform_spacing.
 
     fix_origin : None (defaut) or 1darray
         The origin in physical units (e.g. mm or um) of the fixed image. If None
@@ -63,9 +65,9 @@ def apply_transform(
         The origin in physical units (e.g. mm or um) of the moving image. If None
         the origin is assumed to be (0, 0, 0, ...)
 
-    interpolate_with_nn : Bool (default: False)
-        If true interpolations are done with Nearest Neighbors. Use if warping
-        segmentation/multi-label data.
+    interpolator : string (default: '1')
+        Which interpolation to use.
+        See bigstream.configure_irm.configure_irm documentation for options
 
     extrapolate_with_nn : Bool (default: False)
         If true extrapolations are done with Nearest Neighbors. Use if warping
@@ -80,17 +82,9 @@ def apply_transform(
     """
 
     # set global number of threads
-    if "LSB_DJOB_NUMPROC" in os.environ:
-        ncores = int(os.environ["LSB_DJOB_NUMPROC"])
-    else:
-        ncores = psutil.cpu_count(logical=False)
+    ncores = ut.get_number_of_cores()
     sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(2*ncores)
 
-    # convert images to sitk objects
-    dtype = fix.dtype
-    fix = sitk.Cast(ut.numpy_to_sitk(fix, fix_spacing, fix_origin), sitk.sitkFloat32)
-    mov = sitk.Cast(ut.numpy_to_sitk(mov, mov_spacing, mov_origin), sitk.sitkFloat32)
-
     # construct transform
     fix_spacing = np.array(fix_spacing)
     if transform_spacing is None: transform_spacing = fix_spacing
@@ -101,12 +95,25 @@ def apply_transform(
     # set up resampler object
     resampler = sitk.ResampleImageFilter()
     resampler.SetNumberOfThreads(2*ncores)
-    resampler.SetReferenceImage(fix)
+
+    # set reference data
+    if isinstance(fix, tuple):
+        dtype = mov.dtype
+        resampler.SetSize(fix[::-1])
+        resampler.SetOutputSpacing(fix_spacing[::-1])
+        if fix_origin is not None:
+            resampler.SetOutputOrigin(fix_origin[::-1])
+    else:
+        dtype = fix.dtype
+        fix = sitk.Cast(ut.numpy_to_sitk(fix, fix_spacing, fix_origin), sitk.sitkFloat32)
+        resampler.SetReferenceImage(fix)
+
+    # set moving image and transform
+    mov = sitk.Cast(ut.numpy_to_sitk(mov, mov_spacing, mov_origin), sitk.sitkFloat32)
     resampler.SetTransform(transform)
 
-    # check for NN interpolation
-    if interpolate_with_nn:
-        resampler.SetInterpolator(sitk.sitkNearestNeighbor)
+    # set interpolator
+    resampler.SetInterpolator(interpolator_switch[interpolator])
 
     # check for NN extrapolation
     if extrapolate_with_nn:
@@ -114,7 +121,7 @@ def apply_transform(
 
     # execute, return as numpy array
     resampled = resampler.Execute(mov)
-    return sitk.GetArrayFromImage(resampled).astype(dtype)
+    return sitk.GetArrayViewFromImage(resampled).astype(dtype)
 
 
 def apply_transform_to_coordinates(
@@ -132,19 +139,19 @@ def apply_transform_to_coordinates(
         The coordinates to move. N such coordinates in d dimensions.
 
     transform_list : list
-        The transforms to apply, in stack order. Elements must be 2d 4x4 arrays
-        (affine transforms) or d + 1 dimension ndarrays (deformations).
+        The transforms to apply, in stack order. Elements must be 3x3 or 4x4
+        arrays (affine transforms) or d + 1 dimension ndarrays (deformations).
 
-    transform_spacing : None (default), 1d array, or tuple of 1d arrays
+    transform_spacing : None (default), 1d array, or tuple (not list) of 1d arrays
         The spacing in physical units (e.g. mm or um) between voxels
-        of any deformations in transform_list. If any transform_list
+        of any deformations in transform_list. If transform_list
         contains any deformations then transform_spacing cannot be
         None. If a single 1d array then all deforms have that spacing.
         If a tuple, then its length must be the same as transform_list,
         thus each deformation can be given its own spacing. Spacings
         given for affine transforms are ignored.
 
-    transform_origin : None (default), 1d array, or tuple of 1d arrays
+    transform_origin : None (default), 1d array, or tuple (not list) of 1d arrays
         The origin in physical units (e.g. mm or um) of the given transforms.
         If None, all origins are assumed to be (0, 0, 0, ...); otherwise, follows
         the same logic as transform_spacing. Origins given for affine transforms
@@ -160,10 +167,11 @@ def apply_transform_to_coordinates(
     for iii, transform in enumerate(transform_list[::-1]):
 
         # if transform is an affine matrix
-        if transform.shape == (4, 4):
+        if len(transform.shape) == 2:
 
             # matrix vector multiply
-            mm, tt = transform[:3, :3], transform[:3, -1]
+            ndims = transform.shape[0] - 1
+            mm, tt = transform[:ndims, :ndims], transform[:ndims, -1]
             coordinates = np.einsum('...ij,...j->...i', mm, coordinates) + tt
 
         # if transform is a deformation vector field
@@ -183,10 +191,11 @@ def apply_transform_to_coordinates(
             # get coordinates in transform voxel units, reformat for map_coordinates
             if origin is not None: coordinates -= origin
             coordinates = ( coordinates / spacing ).transpose()
-    
+
             # interpolate position field at coordinates, reformat, return
-            interp = lambda x: map_coordinates(x, coordinates, order=1, mode='nearest')
-            dX = np.array([interp(transform[..., i]) for i in range(3)]).transpose()
+            ndims = transform.shape[-1]
+            interp = lambda x: map_coordinates(x, coordinates, mode='nearest')
+            dX = np.array([interp(transform[..., i]) for i in range(ndims)]).transpose()
             coordinates = coordinates.transpose() * spacing + dX
             if origin is not None: coordinates += origin
 
@@ -196,7 +205,8 @@ def apply_transform_to_coordinates(
 def compose_displacement_vector_fields(
     first_field,
     second_field,
-    spacing,
+    first_spacing,
+    second_spacing,
 ):
     """
     Compose two displacement vector fields into a single field
@@ -209,25 +219,30 @@ def compose_displacement_vector_fields(
     second_field : nd-array
         The second field
 
-    spacing : 1d-array
-        The voxel spacing for the two fields (fields must have same spacing)
+    first_spacing : 1d-array
+        The voxel spacing for the first field
+
+    second_spacing : 1d-array
+        The voxel spacing for the second field
 
     Returns
     -------
     composite_field : nd-array
         The single field composition of first_field and second_field
+        The second field is the one learned second. Thus, the composition
+        is going to be on the same voxel grid and spacing as the second field.
     """
 
     # container for warped first field
-    first_field_warped = np.empty_like(first_field)
+    first_field_warped = np.empty_like(second_field)
 
     # loop over components
-    for iii in range(3):
+    for iii in range(len(first_spacing)):
 
         # warp first field with second
         first_field_warped[..., iii] = apply_transform(
-            first_field[..., iii], first_field[..., iii],
-            spacing, spacing,
+            second_field[..., iii], first_field[..., iii],
+            second_spacing, first_spacing,
             transform_list=[second_field,],
             extrapolate_with_nn=True,
         )
@@ -236,85 +251,106 @@ def compose_displacement_vector_fields(
     return first_field_warped + second_field
 
 
-def compose_transforms(transform_one, transform_two, spacing):
+def compose_transforms(
+    first_transform,
+    second_transform,
+    first_spacing,
+    second_spacing,
+):
     """
     Compose two transforms into a single transform
 
     Parameters
     ----------
-    transform_one : nd-array
-        Can be either a 4x4 affine matrix or a displacement vector field
+    first_transform : nd-array
+        Can be either a 3x3 or 4x4 affine matrix or a displacement vector field
 
-    transform_two : nd-array
-        Can be either a 4x4 affine matrix or a displacement vector field
+    second_transform : nd-array
+        Can be either a 3x3 or 4x4 affine matrix or a displacement vector field
 
-    spacing : 1d-array
-        The voxel spacing for the two transforms (transforms must have same spacing)
-        Ignored for affine matrices.
+    first_spacing : 1d-array
+        The voxel spacing for the first transform
+        Ignored for affine transforms (just put in a dummy value)
+
+    second_spacing : 1d-array
+        The voxel spacing for the second transform
+        Ignored for affine transforms (just put in a dummy value)
 
     Returns
     -------
     composite_transform : nd-array
-        The single transform composition of transform_one and transform_two
+        The single transform composition of first_transform and second_transform
         If both given transforms are affine this is a 4x4 matrix. Otherwise,
         it is a displacement vector field.
     """
 
     # two affines
-    if transform_one.shape == (4, 4) and transform_two.shape == (4, 4):
-        return np.matmul(transform_one, transform_two)
+    if len(first_transform.shape) == 2 and len(second_transform.shape) == 2:
+        return np.matmul(first_transform, second_transform)
 
     # one affine, two field
-    elif transform_one.shape == (4, 4):
-        transform_one = ut.matrix_to_displacement_field(
-            transform_one, transform_two.shape[:-1], spacing,
+    elif len(first_transform.shape) == 2:
+        first_transform = ut.matrix_to_displacement_field(
+            first_transform, second_transform.shape[:-1], second_spacing,
         )
+        first_spacing = second_spacing
 
     # one field, two affine
-    elif transform_two.shape == (4, 4):
-        transform_two = ut.matrix_to_displacement_field(
-            transform_two, transform_one.shape[:-1], spacing,
+    elif len(second_transform.shape) == 2:
+        second_transform = ut.matrix_to_displacement_field(
+            second_transform, first_transform.shape[:-1], first_spacing,
         )
+        second_spacing = first_spacing
 
     # compose fields
     return compose_displacement_vector_fields(
-        transform_one, transform_two, spacing,
+        first_transform, second_transform, first_spacing, second_spacing,
     )
 
 
-def compose_transform_list(transforms, spacing):
+def compose_transform_list(transforms, spacings):
     """
     Compose a list of transforms into a single transform
 
     Parameters
     ----------
     transforms : list
-        Elements of list must be either 4x4 affine matrices or displacement
+        Elements of list must be either 3x3 or 4x4 affine matrices or displacement
         vector fields
 
-    spacing : 1d-array
-        The voxel spacing of all transforms in the list (all transforms must
-        have the same spacing). Ignored for affine transforms.
+    spacings : list of 1d-arrays
+        The voxel spacing of all transforms in the list
+        Ignored for affine transforms (just put in a dummy value)
 
     Returns
     -------
     composite_transform : nd-array
         The single transform composition of all elements in transforms.
-        If all transforms are affine, this is a 4x4 matrix. Otherwise,
+        If all transforms are affine, this is a 3x3 or 4x4 matrix. Otherwise,
         it is a displacement vector field.
     """
 
+    # ensure spacings is a list
+    if not isinstance(spacings, list):
+        spacings = [spacings,] * len(transforms)
+
     transform = transforms.pop()
+    transform_spacing = spacings.pop()
     while transforms:
-        transform = compose_transforms(transforms.pop(), transform, spacing)
+        transform = compose_transforms(
+            transforms.pop(), transform,
+            spacings.pop(), transform_spacing,
+        )
     return transform
 
 
 def invert_displacement_vector_field(
     field,
     spacing,
+    step=0.5,
     iterations=10,
-    order=2,
+    sqrt_order=2,
+    sqrt_step=0.5,
     sqrt_iterations=5,
 ):
     """
@@ -328,13 +364,19 @@ def invert_displacement_vector_field(
     spacing : 1d-array
         The physical voxel spacing of the displacement field
 
+    step : float (default: 0.5)
+        The step size used for each iteration of the stationary point algorithm
+
     iterations : scalar int (default: 10)
         The number of stationary point iterations to find inverse. More
         iterations gives a more accurate inverse but takes more time.
 
-    order : scalar int (default: 2)
+    sqrt_order : scalar int (default: 2)
         The number of roots to take before stationary point iterations
 
+    sqrt_step : float (default: 0.5)
+        The step size used for each iteration of the composition square root gradient descent
+
     sqrt_iterations : scalar int (default: 5)
         The number of iterations to find the field composition square root
 
@@ -350,17 +392,18 @@ def invert_displacement_vector_field(
 
     # initialize inverse as negative root
     root = _displacement_field_composition_nth_square_root(
-        field, spacing, order, sqrt_iterations,
+        field, spacing, sqrt_order, sqrt_step, sqrt_iterations,
     )
     inv = - np.copy(root)
 
     # iterate to invert
     for i in range(iterations):
-        inv -= compose_transforms(root, inv, spacing)
+        residual = compose_transforms(root, inv, spacing, spacing)
+        inv -= step * residual
 
     # square-compose inv order times
-    for i in range(order):
-        inv = compose_transforms(inv, inv, spacing)
+    for i in range(sqrt_order):
+        inv = compose_transforms(inv, inv, spacing, spacing)
 
     # return result
     return inv
@@ -370,7 +413,8 @@ def _displacement_field_composition_nth_square_root(
     field,
     spacing,
     order,
-    sqrt_iterations=5,
+    step,
+    iterations,
 ):
     """
     """
@@ -381,7 +425,7 @@ def _displacement_field_composition_nth_square_root(
     # iterate taking square roots
     for i in range(order):
         root = _displacement_field_composition_square_root(
-            root, spacing, iterations=sqrt_iterations,
+            root, spacing, step, iterations,
         )
 
     # return result
@@ -391,7 +435,8 @@ def _displacement_field_composition_nth_square_root(
 def _displacement_field_composition_square_root(
     field,
     spacing,
-    iterations=5,
+    step,
+    iterations,
 ):
     """
     """
@@ -401,10 +446,36 @@ def _displacement_field_composition_square_root(
 
     # iterate
     for i in range(iterations):
-        residual = (field - compose_transforms(root, root, spacing))
-        root += 0.5 * residual
+        residual = (field - compose_transforms(root, root, spacing, spacing))
+        gradient = np.einsum('...ij,...j', displacement_field_jacobian(root, spacing), residual) + residual
+        root += step * gradient
 
     # return result
     return root
 
 
+def displacement_field_jacobian(field, spacing):
+    """
+    """
+
+    # convert to position field
+    grid = tuple(slice(None, x) for x in field.shape[:-1])
+    position_field = np.mgrid[grid].astype(field.dtype)
+    position_field = np.moveaxis(position_field, 0, -1) * spacing + field
+
+    # get jacobian matrices
+    jacobian = np.empty(field.shape[:-1] + (field.shape[-1],)*2, dtype=field.dtype)
+    for iii in range(field.shape[-1]):
+        grad = np.moveaxis( np.array( np.gradient(position_field[..., iii], *spacing) ), 0, -1)
+        jacobian[..., iii, :] = np.ascontiguousarray(grad)
+    return jacobian
+
+
+def displacement_field_jacobian_determinant(field, spacing):
+    """
+    """
+
+    jacobian = displacement_field_jacobian(field, spacing)
+    return np.linalg.det(jacobian)
+
+
