@@ -3,11 +3,13 @@ from scipy.spatial.transform import Rotation
 import SimpleITK as sitk
 import zarr
 from zarr.indexing import BasicIndexer
-from numcodecs import Blosc
 from distributed import Lock
 import glob
+import os , psutil
 import h5py
 from ClusterWrap.decorator import cluster
+from zarr import blosc
+import psutil
 
 
 def skip_sample(image, spacing, ss_spacing):
@@ -96,9 +98,10 @@ def invert_matrix_axes(matrix):
         The same matrix but with the axis order inverted
     """
 
-    corrected = np.eye(4)
-    corrected[:3, :3] = matrix[:3, :3][::-1, ::-1]
-    corrected[:3, -1] = matrix[:3, -1][::-1]
+    ndims = matrix.shape[0] - 1
+    corrected = np.eye(ndims + 1)
+    corrected[:ndims, :ndims] = matrix[:ndims, :ndims][::-1, ::-1]
+    corrected[:ndims, -1] = matrix[:ndims, -1][::-1]
     return corrected
 
 
@@ -120,9 +123,10 @@ def change_affine_matrix_origin(matrix, origin):
         The same affine transform but encoded with respect to the given origin
     """
 
-    tl, tr = np.eye(4), np.eye(4)
+    ndims = matrix.shape[0] - 1
+    tl, tr = np.eye(ndims+1), np.eye(ndims+1)
     origin = np.array(origin)
-    tl[:3, -1], tr[:3, -1] = -origin, origin
+    tl[:ndims, -1], tr[:ndims, -1] = -origin, origin
     return np.matmul(tl, np.matmul(matrix, tr))
 
 
@@ -141,9 +145,10 @@ def affine_transform_to_matrix(transform):
         The same transform as a 4x4 matrix
     """
 
-    matrix = np.eye(4)
-    matrix[:3, :3] = np.array(transform.GetMatrix()).reshape((3,3))
-    matrix[:3, -1] = np.array(transform.GetTranslation())
+    ndims = transform.GetDimension()
+    matrix = np.eye(ndims+1)
+    matrix[:ndims, :ndims] = np.array(transform.GetMatrix()).reshape((ndims,ndims))
+    matrix[:ndims, -1] = np.array(transform.GetTranslation())
     return invert_matrix_axes(matrix)
 
 
@@ -162,10 +167,11 @@ def matrix_to_affine_transform(matrix):
         The same affine but as a sitk.AffineTransform object
     """
 
+    ndims = matrix.shape[0] - 1
     matrix_sitk = invert_matrix_axes(matrix)
-    transform = sitk.AffineTransform(3)
-    transform.SetMatrix(matrix_sitk[:3, :3].flatten())
-    transform.SetTranslation(matrix_sitk[:3, -1].squeeze())
+    transform = sitk.AffineTransform(ndims)
+    transform.SetMatrix(matrix_sitk[:ndims, :ndims].flatten())
+    transform.SetTranslation(matrix_sitk[:ndims, -1].squeeze())
     return transform
 
 
@@ -184,10 +190,11 @@ def matrix_to_euler_transform(matrix):
         The same rigid transform but as a sitk object
     """
 
+    ndims = matrix.shape[0] - 1
     matrix_sitk = invert_matrix_axes(matrix)
-    transform = sitk.Euler3DTransform()
-    transform.SetMatrix(matrix_sitk[:3, :3].flatten())
-    transform.SetTranslation(matrix_sitk[:3, -1].squeeze())
+    transform = sitk.Euler2DTransform() if ndims == 2 else sitk.Euler3DTransform()
+    transform.SetMatrix(matrix_sitk[:ndims, :ndims].flatten())
+    transform.SetTranslation(matrix_sitk[:ndims, -1].squeeze())
     return transform
 
 
@@ -207,11 +214,17 @@ def euler_transform_to_parameters(transform):
         The rigid transform parameters: (rotX, rotY, rotZ, transX, transY, transZ)
     """
 
-    return np.array((transform.GetAngleX(),
-                     transform.GetAngleY(),
-                     transform.GetAngleZ()) +
-                     transform.GetTranslation()
-    )
+    if transform.GetDimension() == 2:
+        return np.array((transform.GetAngle(),) +
+                         transform.GetTranslation(),
+        )
+
+    elif transform.GetDimension() == 3:
+        return np.array((transform.GetAngleX(),
+                         transform.GetAngleY(),
+                         transform.GetAngleZ()) +
+                         transform.GetTranslation()
+        )
 
 
 def parameters_to_euler_transform(params):
@@ -229,13 +242,20 @@ def parameters_to_euler_transform(params):
         A sitk rigid transform object
     """
 
-    transform = sitk.Euler3DTransform()
-    transform.SetRotation(*params[:3])
-    transform.SetTranslation(params[3:])
-    return transform
+    if len(params) == 3:
+        transform = sitk.Euler2DTransform()
+        transform.SetAngle(params[0])
+        transform.SetTranslation(params[1:])
+        return transform
+
+    elif len(params) == 6:
+        transform = sitk.Euler3DTransform()
+        transform.SetRotation(*params[:3])
+        transform.SetTranslation(params[3:])
+        return transform
 
 
-def physical_parameters_to_affine_matrix(params, center):
+def physical_parameters_to_affine_matrix_3d(params, center):
     """
     Convert separate affine transform parameters to an affine matrix
 
@@ -300,11 +320,11 @@ def matrix_to_displacement_field(matrix, shape, spacing=None, centered=False):
     """
 
     if spacing is None: spacing = np.ones(len(shape))
-    nrows, ncols, nstacks = shape
-    grid = np.array(np.mgrid[:nrows, :ncols, :nstacks])
-    grid = grid.transpose(1,2,3,0) * spacing
+    grid = np.array(np.mgrid[tuple(slice(None, x) for x in shape)])
+    grid = np.moveaxis(grid, 0, -1) * spacing
     if centered: grid += 0.5 * spacing
-    mm, tt = matrix[:3, :3], matrix[:3, -1]
+    ndims = matrix.shape[0] - 1
+    mm, tt = matrix[:ndims, :ndims], matrix[:ndims, -1]
     return np.einsum('...ij,...j->...i', mm, grid) + tt - grid
 
 
@@ -351,9 +371,11 @@ def bspline_parameters_to_transform(parameters):
         A sitk.BSplineTransform object
     """
 
-    t = sitk.BSplineTransform(3, 3)
-    t.SetFixedParameters(parameters[:18])
-    t.SetParameters(parameters[18:])
+    # number of fixed parameters depends on dimension, stored in parameters[0]
+    nfp = 10 if parameters[0] == 2 else 18
+    t = sitk.BSplineTransform(parameters[0], 3)
+    t.SetFixedParameters(parameters[1:nfp+1])
+    t.SetParameters(parameters[nfp+1:])
     return t
 
 
@@ -394,7 +416,7 @@ def bspline_to_displacement_field(
         shape[::-1], origin[::-1], spacing[::-1],
         direction[::-1, ::-1].ravel(),
     )
-    return sitk.GetArrayFromImage(df).astype(np.float32)[..., ::-1]
+    return sitk.GetArrayViewFromImage(df).astype(np.float32)[..., ::-1]
 
 
 # TODO: function that takes a numpy array and return transform type
@@ -453,9 +475,17 @@ def transform_list_to_composite_transform(transform_list, spacing=None, origin=N
         All transforms in the given list compressed into a sitk.CompositTransform 
     """
 
-    transform = sitk.CompositeTransform(3)
+    # determine dimension
+    if len(transform_list[0].shape) == 2:
+        ndims = 2 if transform_list[0].shape == (3, 3) else 3
+    elif len(transform_list[0].shape) > 2:
+        ndims = transform_list[0].ndim - 1
+    else:
+        ndims = transform_list[0][0]
+
+    transform = sitk.CompositeTransform(ndims)
     for iii, t in enumerate(transform_list):
-        if t.shape == (4, 4):
+        if t.shape in [(3, 3), (4, 4)]:
             t = matrix_to_affine_transform(t)
         elif len(t.shape) == 1:
             t = bspline_parameters_to_transform(t)
@@ -467,7 +497,15 @@ def transform_list_to_composite_transform(transform_list, spacing=None, origin=N
     return transform
 
 
-def create_zarr(path, shape, chunks, dtype, chunk_locked=False, client=None):
+def create_zarr(
+    path,
+    shape,
+    chunks,
+    dtype,
+    multithreaded=False,
+    chunk_locked=False,
+    client=None,
+):
     """
     Create a new zarr array on disk
 
@@ -497,15 +535,16 @@ def create_zarr(path, shape, chunks, dtype, chunk_locked=False, client=None):
         Reference to the newly created zarr array on disk
     """
 
-    compressor = Blosc(
-        cname='zstd', clevel=4, shuffle=Blosc.BITSHUFFLE,
-    )
+    synchronizer = None
+    if multithreaded:
+        blosc.use_threads = True
+        synchronizer = zarr.ThreadSynchronizer()
     zarr_disk = zarr.open(
         path, 'w',
         shape=shape,
         chunks=chunks,
         dtype=dtype,
-        compressor=compressor,
+        synchronizer=synchronizer,
     )
 
     # this code is currently never used within bigstream
@@ -521,7 +560,7 @@ def create_zarr(path, shape, chunks, dtype, chunk_locked=False, client=None):
         )
 
     return zarr_disk
-    
+
 
 def numpy_to_zarr(array, chunks, path):
     """
@@ -552,6 +591,20 @@ def numpy_to_zarr(array, chunks, path):
         return array
 
 
+def get_number_of_cores():
+    """
+    Get number of physical cores available to the python process.
+    Currently only considers LSF environment variable. If not an LSF
+    cluster job the uses psutil.cpu_count
+    """
+
+    if "LSB_DJOB_NUMPROC" in os.environ:
+        ncores = int(os.environ["LSB_DJOB_NUMPROC"])
+    else:
+        ncores = psutil.cpu_count(logical=False)
+    return ncores
+
+
 @cluster
 def distributed_directory_of_hdf5_to_zarr(
     directory,
